
//Multi-Inno MI0283QT-2 (HX8347D) - http://www.watterott.net

#include <avr/io.h>
#include <util/delay.h>

#define LCD_MIRROR

//#define LCD_MIRROR         //define to mirror the display
#define LCD_WIDTH            (320)
#define LCD_HEIGHT           (240)
#define RGB(r,g,b)           ((((r)&0xF8)<<8)|(((g)&0xFC)<<3)|(((b)&0xF8)>>3)) //5 red | 6 green | 5 blue

#define LCD_ID               (0)
#define LCD_DATA             ((0x72)|(LCD_ID<<2))
#define LCD_REGISTER         ((0x70)|(LCD_ID<<2))

#define SPI_SCK		(1<<7)
#define SPI_MISO	(1<<6)
#define SPI_MOSI	(1<<5)
#define SPI_SS		(1<<4)

#define TOUCH_CS	(1<<0)
#define LCD_RST		(1<<1)
#define LCD_CS		SPI_SS
#define LCD_LED		(1<<3) 	// OC0

#define TOUCH_CS_DISABLE()   PORTB |= TOUCH_CS;
#define TOUCH_CS_ENABLE()    PORTB &= ~TOUCH_CS;
#define LCD_RST_DISABLE()    PORTB |= LCD_RST;  // high
#define LCD_RST_ENABLE()     PORTB &= ~LCD_RST; // low
#define LCD_CS_DISABLE()     PORTB |= LCD_CS;   // high
#define LCD_CS_ENABLE()      PORTB &= ~LCD_CS;  // low
#define LCD_LED_ON()	     PORTB |= LCD_LED;
#define LCD_LED_OFF()	     PORTB &= ~LCD_LED;

#define DDRB_MASK        (SPI_SCK | SPI_MOSI | SPI_SS | TOUCH_CS | LCD_RST | LCD_LED | LCD_CS | SPI_MISO)
#define DDRB_OUT_MASK    (SPI_SCK | SPI_MOSI | SPI_SS | TOUCH_CS | LCD_RST | LCD_LED | LCD_CS)
#define LCD_PORT_INIT()  do { PORTB |= SPI_MISO; LCD_RST_ENABLE(); LCD_CS_DISABLE(); TOUCH_CS_DISABLE(); DDRB = (DDRB & ~DDRB_MASK) | DDRB_OUT_MASK; } while(0)


void lcd_draw(unsigned int color);
void lcd_drawstop(void);
void lcd_drawstart(void);
void lcd_cursor(unsigned int x, unsigned int y);
void lcd_area(unsigned int x0, unsigned int y0,unsigned int x1, unsigned int y1);
void lcd_cmd(unsigned int reg, unsigned int param);
void lcd_data(unsigned int c);
void lcd_reset(void);
static inline void  spi_wait(void) { /*_delay_ms(0.5);*/ }

#if defined(SOFTWARE_SPI)
static void spi_write(unsigned char data)
{
  uint8_t mask;
  
  for(mask = 0x80; mask != 0; mask >>= 1)
  {
    CLK_LOW();
    if(mask & data)
    {
      MOSI_HIGH();
    }
    else
    {
      MOSI_LOW();
    }
    CLK_HIGH();
  }
  CLK_LOW();

  return;
}
#else
static inline void spi_write(unsigned char data)
{
	SPDR = data;
  	while(!(SPSR & (1<<SPIF)))
		;

  	return;
}
#endif


void lcd_clear(unsigned int color)
{
  unsigned int i;

  lcd_area(0, 0, (LCD_WIDTH-1), (LCD_HEIGHT-1));

  lcd_drawstart();
  for(i = (LCD_WIDTH*(LCD_HEIGHT/8)); i != 0; i--)
  {
    lcd_draw(color); //1
    lcd_draw(color); //2
    lcd_draw(color); //3
    lcd_draw(color); //4
    lcd_draw(color); //5
    lcd_draw(color); //6
    lcd_draw(color); //7
    lcd_draw(color); //8
  }
  lcd_drawstop();

  return;
}


void lcd_draw(unsigned int color)
{
  spi_write((color & 0xFF00)>>8);
  spi_write(color & 0xFF);

  return;
}


void lcd_drawstop(void)
{
  spi_wait();
  LCD_CS_DISABLE();

  return;
}


void lcd_drawstart(void)
{
  LCD_CS_ENABLE();
  spi_write(LCD_REGISTER);
  spi_write(0x22);
  spi_wait();
  LCD_CS_DISABLE();

  LCD_CS_ENABLE();
  spi_write(LCD_DATA);

  return;
}


// x0 = start column, x1 = end column; y0 = start row, y1 = end row
void lcd_area(unsigned int x0, unsigned int y0, unsigned int x1, unsigned int y1)
{
  // Set GRAM area
  lcd_cmd(0x02, ((x0 & 0xFF00) >> 8)); //set x0
  lcd_cmd(0x03, ((x0 & 0x00FF) >> 0)); //set x0
//  lcd_cmd(0x02, ((x0 & 0xFF00) >> 8)); //set x0

  lcd_cmd(0x04, ((x1 & 0xFF00) >> 8)); //set x1
  lcd_cmd(0x05, ((x1 & 0x00FF) >> 0)); //set x1
//  lcd_cmd(0x04, ((x1 & 0xFF00) >> 8)); //set x1

  lcd_cmd(0x06, ((y0 & 0xFF00) >> 8)); //set y0
  lcd_cmd(0x07, ((y0 & 0x00FF) >> 0)); //set y0
//  lcd_cmd(0x06, ((y0 & 0xFF00) >> 8)); //set y0

  lcd_cmd(0x08, ((y1 & 0xFF00) >> 8)); //set y1
  lcd_cmd(0x09, ((y1 & 0x00FF) >> 0)); //set y1
//  lcd_cmd(0x08, ((y1 & 0xFF00) >> 8)); //set y1

  return;
}


void lcd_cursor(unsigned int x, unsigned int y)
{
  lcd_area(x, y, x, y);

  return;
}


void lcd_cmd(unsigned int reg, unsigned int param)
{
  LCD_CS_ENABLE();
  spi_write(LCD_REGISTER);
  spi_write(reg);
  spi_wait();
  LCD_CS_DISABLE();

  LCD_CS_ENABLE();
  spi_write(LCD_DATA);
  spi_write(param);
  spi_wait();
  LCD_CS_DISABLE();

  return;
}


void lcd_data(unsigned int c)
{
  LCD_CS_ENABLE();
  spi_write(LCD_DATA);
  spi_write((c & 0xFF00)>>8);
  spi_write(c & 0xFF);
  spi_wait();
  LCD_CS_DISABLE();

  return;
}


void lcd_reset(void)
{
  // CM02.8+HX8347D
  //reset
  LCD_CS_DISABLE();
  LCD_RST_ENABLE();
  _delay_ms(50);
  LCD_RST_DISABLE();
  _delay_ms(50);

  //driving ability
  lcd_cmd(0xEA, 0x0000);  // PTBA[15:8]
  lcd_cmd(0xEB, 0x0020);  // PTBA[7:0]
  lcd_cmd(0xEC, 0x000C);  // STBA[15:8]
  lcd_cmd(0xED, 0x00C4);  // STBA[7:0]
  lcd_cmd(0xE8, 0x0040);  // OPON[7:0]
  lcd_cmd(0xE9, 0x0038);  // OPON1[7:0]
  lcd_cmd(0xF1, 0x0001);  // OTPS1B
  lcd_cmd(0xF2, 0x0010);  // GEN
  lcd_cmd(0x27, 0x00A3);

#ifdef GAMMA2_2
 #ifndef ARRAY_SIZE
   #define ARRAY_SIZE(a)  (sizeof(a)/sizeof(a[0]))
 #endif
  // Gamma 2.2 Setting
  static const unsigned int gamma2_part1[] =
   { 0x00 ,0x00, 0x01, 0x13, 0x10, 0x26, 0x08, 0x51, 0x02, 0x12, 0x18, 0x19, 0x14 };
  static const unsigned int gamma2_part2[] =
   { 0x19, 0x2f, 0x2c, 0x3e, 0x3f, 0x3f, 0x2e, 0x77, 0x0b, 0x06, 0x07, 0x0d, 0x1d, 0xcc, };
  for (unsigned char i = 0; i < ARRAY_SIZE(gamma2_part1); i++ )
      lcd_cmd(0x40+i, gamma2_part1[i]);
  for (unsigned char i = 0; i < ARRAY_SIZE(gamma2_part2); i++ )
      lcd_cmd(0x50+i, gamma2_part2[i]);
#endif
  //power voltage
  lcd_cmd(0x1B, 0x001B); // VRH=4.65V
  lcd_cmd(0x1A, 0x0001); // BT (VGH~15V,VGL~-10V,DDVDH~5V)
  lcd_cmd(0x24, 0x002F); // VMH(VCOM High voltage ~3.2V)
  lcd_cmd(0x25, 0x0057); // VML(VCOM Low voltage -1.2V)

  //****VCOM offset**
  lcd_cmd(0x23, 0x008D); //for flicker adjust //can reload from OTP

  //power on
  lcd_cmd(0x18, 0x0036); //I/P_RADJ,N/P_RADJ, Normal mode 60Hz
  lcd_cmd(0x19, 0x0001); //OSC_EN='1',start osc
  lcd_cmd(0x01, 0x0000); //DP_STB='0', out deep sleep = wakeup
  lcd_cmd(0x1F, 0x0088); // GAS=1, VOMG=00, PON=0, DK=1, XDK=0, DVDH_TRI=0, STB=0
  _delay_ms(5);
  lcd_cmd(0x1F, 0x0080); // GAS=1, VOMG=00, PON=0, DK=0, XDK=0, DVDH_TRI=0, STB=0
  _delay_ms(5);
  lcd_cmd(0x1F, 0x0090);  // GAS=1, VOMG=00, PON=1, DK=0, XDK=0, DVDH_TRI=0, STB=0
  _delay_ms(5);
  lcd_cmd(0x1F, 0x00D0);  // GAS=1, VOMG=10, PON=1, DK=0, XDK=0, DVDH_TRI=0, STB=0
  _delay_ms(5);

  //color selection
  lcd_cmd(0x17, 0x0005); //0x0005=65k, 0x0006=262k

  //panel characteristic
  lcd_cmd(0x36, 0x0000); // SS_P, GS_P, REV_P, BGR_P

  //display on
  lcd_cmd(0x28, 0x0038); // GON=1, DTE=1, D=1000
  _delay_ms(40);
  lcd_cmd(0x28, 0x003C); // GON=1, DTE=1, D=1100

  //display options
#ifdef LCD_MIRROR
  lcd_cmd(0x16, 0x0068); //MY=0 MX=1 MV=1 ML=0 BGR=1
#else
  lcd_cmd(0x16, 0x00A8); //MY=1 MX=0 MV=1 ML=0 BGR=1
#endif

  lcd_area(0, 0, (LCD_WIDTH-1), (LCD_HEIGHT-1));

  return;
}

#define clock_div	2

int main(void)
{
	LCD_PORT_INIT();
  	//init hardware spi
        // MSB first, so DORD = 0;
	// leading edge is rising, so CPOL = 0;
	// sample on leading edge, so CPHA = 0;
  	switch(clock_div)
  	{
    	case 2:
      		SPCR = (1<<SPE)|(1<<MSTR); //enable SPI, Master, clk=Fcpu/4
      		SPSR = (1<<SPI2X); //clk*2 = Fcpu/2
      		break;
    	case 4:
      		SPCR = (1<<SPE)|(1<<MSTR); //enable SPI, Master, clk=Fcpu/4
      		SPSR = (0<<SPI2X); //clk*2 = off
      		break;
    	case 8:
      		SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR0); //enable SPI, Master, clk=Fcpu/16
      		SPSR = (1<<SPI2X); //clk*2 = Fcpu/8
      		break;
    	case 16:
      		SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR0); //enable SPI, Master, clk=Fcpu/16
      		SPSR = (0<<SPI2X); //clk*2 = off
      		break;
    	case 32:
      		SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR1); //enable SPI, Master, clk=Fcpu/64
      		SPSR = (1<<SPI2X); //clk*2 = Fcpu/32
      		break;
    	case 64:
      		SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR1); //enable SPI, Master, clk=Fcpu/64
      		SPSR = (0<<SPI2X); //clk*2 = off
      		break;
    	case 128:
      		SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR1)|(1<<SPR0); //enable SPI, Master, clk=Fcpu/128
      		SPSR = (0<<SPI2X); //clk*2 = off
      		break;
  	}

	lcd_reset();
	LCD_LED_ON();

	while(1)
	{
		unsigned char i;

		lcd_clear(RGB(255,255,255));
		_delay_ms(555);
		lcd_clear(RGB(0,255,255));
		_delay_ms(555);
		lcd_clear(RGB(0,0,255));
		_delay_ms(555);
		lcd_clear(RGB(0,0,0));
		_delay_ms(555);
		lcd_clear(RGB(255,255,0));
		_delay_ms(555);
		lcd_clear(RGB(255,0,255));
		_delay_ms(555);
		lcd_clear(RGB(255,0,0));
		_delay_ms(555);
		lcd_clear(RGB(0,255,0));
		_delay_ms(555);
		i = 255;
		while( i > 3)
		{
			lcd_clear(RGB(i,0,0)); i+= 3;
			_delay_ms(2);
		}
		i = 255;
		while( i > 3)
		{
			lcd_clear(RGB(i,i,i));
			_delay_ms(2);
			i += 3;
		}
		i = 255;
		while( i > 2)
		{
			lcd_clear(RGB(0,i,0)); i += 2;
			_delay_ms(1.5);
		}
		i = 255;
		while( i > 3)
		{			
			lcd_clear(RGB(i,i,0)); i+=3;
			_delay_ms(2);
		}
		i = 255;
		while( i > 3)
		{
			lcd_clear(RGB(i,0,i)); i+=3;
			_delay_ms(2);
		}
		i = 255;
		while( i > 3)
		{
			lcd_clear(RGB(255-i,255-i,i)); i+=3;
			_delay_ms(2);
		}
	}
	return 0;	
}
